/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 */

/*
 * Copyright (c) 2001, 2006 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 *
 * Copyright (c) 2001, 2002, 2006 by
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 */

/*!
    \page lcmaps_gums.mod GUMS plugin

    \section posixsynopsis SYNOPSIS
    \b lcmaps_gums.mod

    \section posixdesc DESCRIPTION

The GUMS module will connect to the GUMS server using the SAML protocol (provided by the Privilege Project) and return a Unix UID, GID and possibly secondary GIDs to the framework.


*/


/*!
    \file   lcmaps_gums.c
    \brief  Plugin to connect to a GUMS server and return a UID, a GID and secondary GIDs.
    \author Oscar Koeroo for the EGEE.
*/





/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <netdb.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
#include <lcmaps/lcmaps_vo_data.h>

#include "lcmaps_gums.h"

/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


// What do we want to do with this time presentation?
#define GMT               1
#define LOCAL             2
#define LEAVE_TIME_ITS_OK 3


/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);


/******************************************************************************
                       Define module specific variables
******************************************************************************/
/*
 * extern char **environ;
 */

static char *execname   = NULL;
static char **exec_argv = NULL;
static int  num_of_args = 0;

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{
  char * logstr = "\tlcmaps_plugin_gums-plugin_initialize()";
  int i;

  for (i=0; i<argc; i++) {
    lcmaps_log_debug(2, "%s: arg %d is %s\n", logstr, i, argv[i]);

    if ((strncasecmp(argv[i], PLUGIN_GUMS_EXEC, sizeof(PLUGIN_GUMS_EXEC)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
        execname = strdup(argv[++i]);
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GUMS_EXEC, execname);
      }
    }
      else if ((strncasecmp(argv[i], PLUGIN_GUMS_ARG, sizeof(PLUGIN_GUMS_ARG)) == 0) && ((i+1) < argc)) {

        if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
        num_of_args++;

        if (exec_argv == NULL) {
          exec_argv = (char *)malloc(sizeof(char *)*2);   //  Allocate the terminating NULL pointer also!
        } else {
          exec_argv = realloc(exec_argv, sizeof(char *) * (num_of_args + 1));
        }

        exec_argv[num_of_args-1] = strdup(argv[i+1]);
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GUMS_ARG, exec_argv[num_of_args-1]);
      }
    }
  }

  lcmaps_log_debug(1, "%s: Initialization succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  char *logstr = "\tlcmaps_plugin_gums-plugin_introspect()";
  static lcmaps_argument_t argList[] = {
    {"pem_string", "char *" ,  1, NULL},
    {NULL        ,  NULL    , -1, NULL}
  };

  lcmaps_log_debug(2, "%s: introspecting\n", logstr);

  *argv = argList;
  *argc = lcmaps_cntArgs(argList);
  lcmaps_log_debug(1, "%s: address first argument: 0x%x\n", logstr, argList);

  lcmaps_log_debug(1, "%s: Introspect succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;
}




/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(int argc, lcmaps_argument_t * argv)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(int argc, lcmaps_argument_t * argv)
{
  return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv, int lcmaps_mode)
{
  FILE *fh = NULL;
  char *logstr = NULL;
  char *datetime = NULL;
  char *encoded_dn = "(void)";
  int p2c[2], c2p[2];
  time_t clock;
  struct tm *tmpTime;
  int rc,rc2,rc3;
  int uid, gid;
#define MAX_SGIDS 128
  int s_gids[MAX_SGIDS];
  int s_rc,s_nr_gids,i;
  int status;
  int exit_code = -1;
  pid_t pid;
  char *pem_string;

  struct passwd * user_info = NULL;
  struct group  * gr_info   = NULL;


  time(&clock);
  tmpTime = gmtime(&clock);
  datetime = malloc(sizeof(char) * 20);
  snprintf(datetime, 20, "%04d-%02d-%02d %02d:%02d:%02d",
           tmpTime->tm_year + 1900, tmpTime->tm_mon + 1, tmpTime->tm_mday,
           tmpTime->tm_hour, tmpTime->tm_min, tmpTime->tm_sec);

  if (lcmaps_mode == PLUGIN_RUN)
    logstr = "\tlcmaps_plugin_gums-plugin_run()";
  else if (lcmaps_mode == PLUGIN_VERIFY)
    logstr = "\tlcmaps_plugin_gums-plugin_verify()";
  else {
    lcmaps_log(0, "\tlcmaps_plugin_gums-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n", lcmaps_mode);

    goto gums_failure;
    /* return(lcmaps_plugin_gums_failed(logstr, datetime)); */
  }

  if ((pem_string = *(char **) lcmaps_getArgValue("pem_string", "char *", argc, argv))) {
    lcmaps_log_debug(1, "%s: found PEM string\n", logstr);
  } else {
    lcmaps_log(0, "%s: could not get value of PEM string.\n", logstr);

    goto gums_failure;
    /* return(lcmaps_plugin_gums_failed(logstr, datetime)); */
  }

  pipe(p2c);
  pipe(c2p);

  pid = fork();
  switch(pid) {
    //  Child Process
    case 0:
      close(p2c[1]);
      close(c2p[0]);
      dup2(p2c[0], 0);
      dup2(c2p[1], 1);
      close(p2c[0]);
      close(c2p[1]);

      //  Setting the real and effective uid/gid to root.
      if (setreuid(0, 0) != 0) {
        lcmaps_log_debug(0, "%s: Setting the real/effective uid to root failed.\n", logstr);
      }

      if (setregid(0, 0) != 0) {
        lcmaps_log_debug(0, "%s: Setting the real/effective gid to root failed.\n", logstr);
      }

      execve(execname, exec_argv, NULL);

      //  Above function should not return. If it does, there is an error.
      goto gums_failure;

      /* return(lcmaps_plugin_gums_failed(logstr, datetime)); */
      break;

    //  Error
    case -1:
      goto gums_failure;
      /* return(lcmaps_plugin_gums_failed(logstr, datetime)); */
      break;

    //  Parent process
    default:
      close(p2c[0]);
      close(c2p[1]);
  }

  //  Dump the string to stdout first and close it when finished.
  write(p2c[1], pem_string, strlen(pem_string));
  close(p2c[1]);

  //  After dumping the string to stdout, we need to expect an answer
  //  back on stdin.
  fh = fdopen(c2p[0], "r");
  rc = fscanf(fh, "GUMSRET UID %d GID %d", &uid, &gid);

  // Added by I. Sfiligoi
  s_nr_gids=0;
  while (1) {
   s_rc=fscanf(fh, " SGID %d", &(s_gids[s_nr_gids]));
   if (s_rc<1) break; // no more SGIDs
   s_nr_gids++;
   if (s_nr_gids>=MAX_SGIDS) break; // reached max
  }

  // this is the terminating string
  // rc3 must be 0
  rc2 = fscanf(fh, " GUMSEND %d", &rc3);

  fclose(fh);

  wait(&status);

  exit_code = WEXITSTATUS(status);

  if (exit_code != 0)
  {
     lcmaps_log (0, "Error: failure in execute. Exit code %d. Executed \"%s\".\n", exit_code, execname);

     /* failure in child */
     goto gums_failure;
  }

  /*
   * Igor's third issue:
   *    fscanf can return plain jabber not related to any UID GID combination.
   *    Need to filter
   */
  if (rc!=2) {
     lcmaps_log (0, "%s Error: invalid return. Expected at least 2 numbers, got only %d\n", logstr,rc);
     goto gums_failure;
  }

  if ((rc2!=1) || (rc3!=0)) {
     lcmaps_log (0, "%s Error: invalid return. Missing terminator.\n", logstr);
     goto gums_failure;
  }

  if (s_nr_gids>=MAX_SGIDS) {
     lcmaps_log (0, "%s Error: Too many SGIDs.  %d>=%d\n", logstr,s_nr_gids,MAX_SGIDS);
     goto gums_failure;
  }

  user_info = getpwuid(uid);
  if (user_info == NULL)
  {
     lcmaps_log (0, "%s Error: User ID doesn't exist on the system: %d\n", logstr, uid);
     goto gums_failure;
     /* return(lcmaps_plugin_gums_failed(logstr, datetime)); */
  }

/*
  gr_info = getgrgid(gid);
  if (gr_info != NULL)
  {
     lcmaps_log (0, "Group ID doesn't exist on the system: %d\n", gid);
     return(lcmaps_plugin_gums_failed(logstr, datetime));
  }
*/


  lcmaps_log (5, "%s: Adding UID :  %d\n", logstr, uid);
  rc=addCredentialData(UID, &uid);
  lcmaps_log (5, "%s: Adding PRI_GID :  %d\n", logstr, gid);
  rc+=addCredentialData(PRI_GID, &gid);

  for (i=0; i<s_nr_gids; i++) {
    lcmaps_log (5, "%s: Adding SEC_GID :  %d\n", logstr, s_gids[i]);
    rc+=addCredentialData(SEC_GID,&(s_gids[i]));
  }

  //  Add the pool index by hand. Normally the poolindex plug in takes
  //  care of this. Since the GUMS plugin setup the poolaccounts, we need
  //  to have a 'NULL' account inorder to prevent LCMAPS from complaining
  //  about missing poolaccounts.
  lcmaps_log (5, "%s: Adding POOL_INDEX :  %s\n", logstr, encoded_dn);
  rc+=addCredentialData(POOL_INDEX, &encoded_dn);

  if (rc==0) {
     lcmaps_log_time(0, "%s: gums plugin succeeded\n", logstr);
  } else {
     lcmaps_log (0, "%s Error: At least one of the addCredentialData calls failed.\n", logstr);
     goto gums_failure;
  }

  if (datetime) free (datetime);
  return LCMAPS_MOD_SUCCESS;


gums_failure:
  if (datetime) free (datetime);
  lcmaps_log_time(0, "%s: gums plugin failed\n", logstr);

  return LCMAPS_MOD_FAIL;
}


int lcmaps_plugin_gums_failed(const char *logstr, char **datetime)
{
  lcmaps_log_debug(1, "%s: done\n", logstr);

  // Free stuff
  if (*datetime) {
    free(*datetime);
    *datetime = NULL;
  }

  lcmaps_log_time(0, "%s: gums plugin failed\n", logstr);

  return LCMAPS_MOD_FAIL;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    char * logstr = "\tlcmaps_plugin_gums-plugin_terminate()";
    lcmaps_log_debug(1,"%s: terminating\n", logstr);


    return LCMAPS_MOD_SUCCESS;
}
